#!/usr/bin/python

from __future__ import print_function

__docformat__ = 'restructedtext en'

import six.moves.cPickle as pickle
import gzip
import os
import sys
import timeit
import numpy
import math
import theano
import theano.tensor as T
from theano.tensor.nnet import conv2d
from theano.tensor.signal import pool

from FunctionList import *

def RunCNN(learning_rate=0.01, n_epochs=1000, operation_string="3,1,1,0,4", dataset='../Data/H1cmplx_lowres_img.sav', batch_size=20, nkerns=(20,50,90,500,100), filter_size=(0-5-5,0-5-5,0-5-5,0-0-0,0-0-0), pool_size=(0-2-2,0-2-2,0-2-2,0-0-0,0-0-0), datacutratio=1, input_dim_x=65, input_dim_y=61, input_dim_d=2, functions=(T.tanh,T.tanh,T.tanh,T.tanh,T.tanh), parameter=(None,None,None,None,None), patience=10000, patience_increase=2, improvement_threshold=0.995):

    datasets = load_data(dataset)

    train_set_xA, train_set_yA = datasets[0]
    valid_set_xA, valid_set_yA = datasets[1]
    test_set_xA, test_set_yA = datasets[2]
    
    if int(operation_string.split(",")[-1])==40:
        train_set_y = theano.shared(numpy.squeeze(numpy.array(train_set_yA.eval(), dtype='f')))
        valid_set_y = theano.shared(numpy.squeeze(numpy.array(valid_set_yA.eval(), dtype='f')))
        test_set_y = theano.shared(numpy.squeeze(numpy.array(test_set_yA.eval(), dtype='f')))
    elif int(operation_string.split(",")[-1])==4:
        train_set_y = theano.shared(numpy.squeeze(numpy.array(train_set_yA.eval(), dtype='int32')))
        valid_set_y = theano.shared(numpy.squeeze(numpy.array(valid_set_yA.eval(), dtype='int32')))
        test_set_y = theano.shared(numpy.squeeze(numpy.array(test_set_yA.eval(), dtype='int32')))
    else:
	print("Error: operations do not end with a regression command.")
        exit(0)


#    print(train_set_xA.eval().shape,valid_set_xA.eval().shape,test_set_xA.eval().shape)
#    exit(0)

    Ntraindata = int(train_set_xA.get_value(borrow=True).shape[0])
    Nvaliddata = int(valid_set_xA.get_value(borrow=True).shape[0])
    Ntestdata = int(test_set_xA.get_value(borrow=True).shape[0])

    if True:
        train_set_x = theano.shared(numpy.expand_dims(train_set_xA.get_value(borrow=True), axis=2).reshape((Ntraindata,input_dim_d,1,input_dim_x,input_dim_y)))
        valid_set_x = theano.shared(numpy.expand_dims(valid_set_xA.get_value(borrow=True), axis=2).reshape((Nvaliddata,input_dim_d,1,input_dim_x,input_dim_y)))
        test_set_x = theano.shared(numpy.expand_dims(test_set_xA.get_value(borrow=True), axis=2).reshape((Ntestdata,input_dim_d,1,input_dim_x,input_dim_y)))
    
    else:
        print("Error: input depth does not match data depth.")
        exit(0)

    # compute number of minibatches for training, validation and testing
    n_train_batches = int(train_set_x.get_value(borrow=True).shape[0] * datacutratio // batch_size)
    n_valid_batches = int(valid_set_x.get_value(borrow=True).shape[0] * datacutratio // batch_size)
    n_test_batches = int(test_set_x.get_value(borrow=True).shape[0] * datacutratio // batch_size)

    ######################
    # BUILD ACTUAL MODEL #
    ######################
    print('... building the model')

    dtensor5 =  T.TensorType('float32', (False,)*5)

    index = T.lscalar()  # index to a [mini]batch
    x = dtensor5('x')   # the data is presented as rasterized images

    if int(operation_string.split(",")[-1])==40:
        y = T.vector('y', dtype='float32')  # the labels are presented as 1D vector of float labels
    elif int(operation_string.split(",")[-1])==4:
        y = T.ivector('y')  # the labels are presented as 1D vector of int labels
    else:
        print("Error: operations do not end with a regression command.")
        exit(0)

    layer=[0]*len(operation_string.split(","))
    layer_input=[0]*len(operation_string.split(","))

    rng = numpy.random.RandomState(1234)

    input = x.reshape((batch_size, input_dim_d, 1, input_dim_x, input_dim_y))

    nkern = 1

    dim_d = input_dim_d
    dim_x = input_dim_x
    dim_y = input_dim_y

    dict_functions={}
    dict_functions["tanh"]=T.tanh
    dict_functions["sigmoid"]=T.nnet.sigmoid
    dict_functions["relu"]=T.nnet.relu
    dict_functions["None"]=None

    for i,op in enumerate(operation_string.split(",")):

        if int(op)==1:
            layer_input[i]=input.reshape((batch_size, dim_d, nkern, dim_x, dim_y))
            layer[i] = LeNetConvPool3DLayer(
                rng,
                input=layer_input[i],
                image_shape=(batch_size, dim_d, nkern, dim_x, dim_y),
                filter_shape=(nkerns[i], filter_size[i][0], nkern, filter_size[i][1], filter_size[i][2]),
                poolsize=(pool_size[i][1], pool_size[i][0], pool_size[i][2]),
                activation=dict_functions[functions[i]],
                reduce_shape=False,
                parameters=parameter[i]
            )
            dim_x = Size(dim_x, filter_size[i][1], pool_size[i][1])
            dim_y = Size(dim_y, filter_size[i][2], pool_size[i][2])
            dim_d = Size(dim_d, filter_size[i][0], pool_size[i][0])
            print("... convolution layer "+str(i)+" created")

        elif int(op)==0:
#            layer_input[i] = input.dimshuffle(0,2,1,3,4).flatten(2)
            layer_input[i] = input.flatten(2)
            layer[i] = HiddenLayer(
                rng,
                input=layer_input[i],
                n_in=nkern * dim_x * dim_y * dim_d,
                n_out=nkerns[i],
                activation=dict_functions[functions[i]],
                parameters=parameter[i]
            )
            print("... hidden layer "+str(i)+" created")

        elif int(op)==4:
            layer_input[i]=input.reshape((batch_size, nkern))
            layer[i] = LogisticRegression(
                input=layer_input[i],
                n_in=int(nkern),
                n_out=int(nkerns[i]),
                parameters=parameter[i]
            )        
            print("... Logistic regression layer created")
            break
   
        elif int(op)==40:
            layer_input[i]=input.reshape((batch_size, nkern))
            layer[i] = MSERegression(
                input=layer_input[i],
                n_in=int(nkern),
                n_out=int(nkerns[i]),
                parameters=parameter[i]
            )
            print("... MSE regression layer created")
            break

        else:
            print("Error: Unknown operation input: "+str(op))
            exit(0)

        nkern = nkerns[i]
        input = layer[i].output
  
        

    # the cost we minimize during training is the NLL of the model

    if int(operation_string.split(",")[-1])==4:
        cost=layer[-1].negative_log_likelihood(y)
        error=layer[-1].errors(y)

    elif int(operation_string.split(",")[-1])==40:
        cost=layer[-1].mean_squared_error(y)
        error=T.sqrt(cost)

    else:
        print("Error: operations do not end with a regression command.")
        exit(0)

    # create a function to compute the mistakes that are made by the model
    test_model = theano.function(
        [index],
        error,
        givens={
            x: test_set_x[index * batch_size: (index + 1) * batch_size],
            y: test_set_y[index * batch_size: (index + 1) * batch_size]
        }
    )

    validate_model = theano.function(
        [index],
        error,
        givens={
            x: valid_set_x[index * batch_size: (index + 1) * batch_size],
            y: valid_set_y[index * batch_size: (index + 1) * batch_size]
        }
    )

    # create a list of all model parameters to be fit by gradient descent

    params = []

    for i,op in enumerate(operation_string.split(",")):
        params+=layer[i].params

    # create a list of gradients for all model parameters
    grads = T.grad(cost, params)

    updates = [
        (param_i, param_i - learning_rate * grad_i)
        for param_i, grad_i in zip(params, grads)
    ]

    train_model = theano.function(
        [index],
        cost,
        updates=updates,
        givens={
            x: train_set_x[index * batch_size: (index + 1) * batch_size],
            y: train_set_y[index * batch_size: (index + 1) * batch_size]
        }
    )

#    print(train_model[6])
#    print(train_set_y.eval()[:6])
#    exit()

 # end-snippet-5

    ###############
    # TRAIN MODEL #
    ###############
    print('... training')

    # early-stopping parameters

    validation_frequency = min(n_train_batches, patience // 2)
                                  # go through this many
                                  # minibatche before checking the network
                                  # on the validation set; in this case we
                                  # check every epoch

    best_validation_loss = numpy.inf
    best_iter = 0
    test_score = 0.
    start_time = timeit.default_timer()

    epoch = 0
    done_looping = False

    while (epoch < n_epochs) and (not done_looping):
        epoch = epoch + 1
        for minibatch_index in range(n_train_batches):
            minibatch_avg_cost = train_model(minibatch_index)
            # iteration number
            iter = (epoch - 1) * n_train_batches + minibatch_index

            if (iter + 1) % validation_frequency == 0:
                # compute zero-one loss on validation set
                validation_losses = [validate_model(i) for i
                                     in range(n_valid_batches)]
                this_validation_loss = numpy.mean(validation_losses)

                if int(operation_string.split(",")[-1])==40:
                    print(
                        'epoch %i, minibatch %i/%i, estimation RMS on validation set %f' %
                        (
                            epoch,
                            minibatch_index + 1,
                            n_train_batches, 
                            this_validation_loss
                        )
                    )

                elif int(operation_string.split(",")[-1])==4:
                    print(
                        'epoch %i, minibatch %i/%i, validation error %f %%' %
                        (
                            epoch,
                            minibatch_index + 1,
                            n_train_batches,
                            this_validation_loss * 100.
                        )
                    )

                else:
              	    print("Error: operations do not end with a regression command.")
                    exit(0)
           
                # if we got the best validation score until now
                if this_validation_loss < best_validation_loss:
                    #improve patience if loss improvement is good enough
                    if (
                        this_validation_loss < best_validation_loss *
                        improvement_threshold
                    ):
                        patience = max(patience, iter * patience_increase)

                    best_validation_loss = this_validation_loss
                    best_iter = iter

                    # test it on the test set
                    test_losses = [test_model(i) for i
                                   in range(n_test_batches)]
                    test_score = numpy.mean(test_losses)

                    if int(operation_string.split(",")[-1])==40:
                        print(('     epoch %i, minibatch %i/%i, RMS of '
                              'best model on test set %f') %
                              (epoch, minibatch_index + 1, n_train_batches,
                              test_score))

                    elif int(operation_string.split(",")[-1])==4:
                        print(('     epoch %i, minibatch %i/%i, test error of '
                               'best model %f %%') %
                               (epoch, minibatch_index + 1, n_train_batches,
                               test_score * 100.))

                    else:
                        print("Error: operations do not end with a regression command.")
                        exit(0)

                    # save the best model
		    for i,op in enumerate(operation_string.split(",")):
                        with open('layer_'+str(i)+'.pkl','wb') as f:
			    pickle.dump(layer[i],f)

            if patience <= iter:
                done_looping = True
                break

    end_time = timeit.default_timer()
    print('For file '+str(dataset)+', with a learning rate: '+str(learning_rate)+', Epochs: '+str(epoch)+', Op string: '+str(operation_string)+', batch size: '+str(batch_size)+', nkerns: '+str(nkerns)+', filter size: '+str(filter_size)+', max pool size: '+str(pool_size)+', data cut ratio: '+str(datacutratio))
    if int(operation_string.split(",")[-1])==40:
        print(('Optimization complete. Best validation RMS of %f'
              ' obtained at iteration %i, with test RMS %f') %
             (best_validation_loss, best_iter + 1, test_score))
    if int(operation_string.split(",")[-1])==4:
        print(('Optimization complete. Best validation score of %f %% '
               'obtained at iteration %i, with test performance %f %%') %
              (best_validation_loss * 100., best_iter + 1, test_score * 100.))
    print(('The code for file ' +
           os.path.split(__file__)[1] +
           ' ran for %.2fm' % ((end_time - start_time) / 60.)), file=sys.stderr)



filter_size=numpy.array([i.split("-") for i in numpy.array(sys.argv[7].split(","))]).astype('int')
pool_size=numpy.array([i.split("-") for i in numpy.array(sys.argv[8].split(","))]).astype('int')

if True:
    if CorrectSizeCheck(operation_string=sys.argv[3], filter_size=numpy.array([i.split("-") for i in numpy.array(sys.argv[7].split(","))]).astype('int'), pool_size=numpy.array([i.split("-") for i in numpy.array(sys.argv[8].split(","))]).astype('int'), input_dim_x=int(sys.argv[10]), input_dim_y=int(sys.argv[11]), input_dim_d=int(sys.argv[12])):
        RunCNN(learning_rate=float(sys.argv[1]), n_epochs=int(sys.argv[2]), operation_string=sys.argv[3], dataset=sys.argv[4], batch_size=int(sys.argv[5]), nkerns=numpy.array(sys.argv[6].split(",")).astype('int'), filter_size=numpy.array([i.split("-") for i in numpy.array(sys.argv[7].split(","))]).astype('int'), pool_size=numpy.array([i.split("-") for i in numpy.array(sys.argv[8].split(","))]).astype('int'), datacutratio=float(sys.argv[9]), input_dim_x=int(sys.argv[10]), input_dim_y=int(sys.argv[11]), input_dim_d=int(sys.argv[12]), functions=numpy.array(sys.argv[13].split(",")), parameter=numpy.array(sys.argv[14].split(",")), patience=int(sys.argv[15]), patience_increase=int(sys.argv[16]), improvement_threshold=float(sys.argv[17]))
        if int(sys.argv[3].split(",")[-1])==40:
            My_MSEPrediction(sys.argv[3].split(","), numpy.array(sys.argv[6].split(",")).astype('int'), sys.argv[4], int(sys.argv[5]), int(sys.argv[12]), numpy.array([i.split("-") for i in numpy.array(sys.argv[7].split(","))]).astype('int'), int(sys.argv[10]), int(sys.argv[11]), numpy.array([i.split("-") for i in numpy.array(sys.argv[8].split(","))]).astype('int'))
        elif int(sys.argv[3].split(",")[-1])==4:
            My_Prediction(sys.argv[3].split(","), numpy.array(sys.argv[6].split(",")).astype('int'), sys.argv[4], int(sys.argv[5]), int(sys.argv[12]), numpy.array([i.split("-") for i in numpy.array(sys.argv[7].split(","))]).astype('int'), int(sys.argv[10]), int(sys.argv[11]), numpy.array([i.split("-") for i in numpy.array(sys.argv[8].split(","))]).astype('int'))
                

    else:
        print('Error: dimensionality is incorrect. Check filter sizes!')
        exit(0)
else:
    print('Error: sizes of operation string or x,y or depth filters do not match!')
    exit(0)


