#!/bin/bash

# Use GPU:

export THEANO_FLAGS="mode=FAST_RUN,device=gpu,floatX=float32,lib.cnmem=0.1,exception_verbosity=high" 

# Location and name of training/validation/test sets:

dataset="../../deepdata/BBH_1s_8192Hz_3K_SNR1p0_z1_ts.sav"

# Input dimensions:

input_dim_x=1		# for time series this is 1		
input_dim_y=8192	# actual time series data
input_dim_d=1		# 3rd dimension for stacked images

# Learning constraints:

learning_rate=0.01
n_epochs=50
batch_size=20
datacutratio=1			# fraction of data you want to use (from start)
patience=100000
patience_increase=2
improvement_threshold=0.995

###########################################
# 'operation_string' operations:
# o convolution layer - 1
# o hidden layer - 0
# o logistic regression layer - 4
# o linear regression layer - 40
###########################################

# CNN hyperparmeters

operation_string="1,1,1,0,4"			# defines layer type
nkerns="16,32,64,64,2"				# the number of neurons in each layer
filter_size="1-1-8,1-1-8,1-1-8,0-0-0,0-0-0"	# filter sizes in each layer (depth,x,y only relevant for conv layers)
pool_size="1-1-4,1-1-4,1-1-4,0-0-0,0-0-0"	# max pool sizes for each layer (depth,x,y only relevant for conv layers) 


# Activation function (tanh, sigmoid, relu, None):

functions="tanh,tanh,tanh,tanh,None"		# activation layers

# Use saved parameters:

parameter="None,None,None,None,None"		# presaved layers that can be loaded in as start points


./CNN.py $learning_rate $n_epochs $operation_string $dataset $batch_size $nkerns $filter_size $pool_size $datacutratio $input_dim_x $input_dim_y $input_dim_d $functions $parameter $patience $patience_increase $improvement_threshold
